/* This file is part of the KDE Project
   Copyright (c) 2008 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "nepomukalignment.h"

#include <Soprano/Client/DBusClient>
#include <Soprano/Client/DBusModel>
#include <Soprano/Node>
#include <Soprano/BindingSet>
#include <Soprano/Statement>
#include <Soprano/StatementIterator>
#include <Soprano/QueryResultIterator>
#include <Soprano/Vocabulary/NAO>

#include <nepomuk/resource.h>

#include <KDebug>
#include <KPluginFactory>

#include <QtCore/QTimer>


Nepomuk::Alignment::Alignment( QObject* parent, const QVariantList& )
    : KDEDModule( parent ),
      m_sopranoClient( 0 ),
      m_sopranoModel( 0 )
{
    // initial run
    // FIXME: At some point this is all way too much for one process or one thread
    QTimer::singleShot( 0, this, SLOT( slotStatementsAdded() ) );
}


Nepomuk::Alignment::~Alignment()
{
}


Soprano::Model* Nepomuk::Alignment::sopranoModel()
{
    if ( !m_sopranoClient ) {
        m_sopranoClient = new Soprano::Client::DBusClient( "org.kde.NepomukServer", this );
    }
    if ( !m_sopranoModel ) {
        if ( m_sopranoModel = m_sopranoClient->createModel( "main" ) ) {
            connect( m_sopranoModel, SIGNAL( statementsAdded() ),
                     this, SLOT( slotStatementsAdded() ) );
        }
    }
    return m_sopranoModel;
}


void Nepomuk::Alignment::slotStatementsAdded()
{
    // FIXME: do plugins
    fixupTagUris();
}


void Nepomuk::Alignment::fixupTagUris()
{
    if ( sopranoModel() ) {
        // all we do here is to ensure that all nao:Tag resources have a uri like "tags:/<label>" so
        // it can be used transparnetly in the tags kioslave
        QList<Soprano::BindingSet> tagsToChange = sopranoModel()->executeQuery( QString( "select distinct ?tag ?name where { "
                                                                                         "?tag a <%1> . "
                                                                                         "?tag <%2> ?name . "
                                                                                         "FILTER(!REGEX(STR(?tag),'^tags:/')) . }" )
                                                                                .arg( Soprano::Vocabulary::NAO::Tag().toString() )
                                                                                .arg( Nepomuk::Resource::labelUri() ),
                                                                                Soprano::Query::QueryLanguageSparql ).allBindings();

        QList<Soprano::BindingSet>::const_iterator end = tagsToChange.constEnd();
        for ( QList<Soprano::BindingSet>::const_iterator it = tagsToChange.constBegin();
              it != end; ++it ) {
            const Soprano::BindingSet& set = *it;

            QUrl oldUri = set["tag"].uri();
            QString name = set["name"].toString();

            if ( !name.isEmpty() ) {
                QUrl newUri = "tags:/" + name;

                kDebug() << "Changing tag uri" << oldUri << "to" << newUri;

                // If only we had SPARQL update queries!
                QList<Soprano::Statement> tagStatements = sopranoModel()->listStatements( oldUri, Soprano::Node(), Soprano::Node() ).allStatements();
                foreach( Soprano::Statement s, tagStatements ) {
                    sopranoModel()->removeStatement( s );
                    s.setSubject( newUri );
                    sopranoModel()->addStatement( s );
                }

                tagStatements = sopranoModel()->listStatements( Soprano::Node(), Soprano::Node(), oldUri ).allStatements();
                foreach( Soprano::Statement s, tagStatements ) {
                    sopranoModel()->removeStatement( s );
                    s.setObject( newUri );
                    sopranoModel()->addStatement( s );
                }
            }
            else {
                kDebug() << "Invalid tag data. Cannot update URI:" << oldUri << name;
            }
        }
    }
    else {
        kDebug() << "Could not contact Nepomuk server.";
    }
}


K_PLUGIN_FACTORY(NepomukAlignmentFactory, registerPlugin<Nepomuk::Alignment>();)
K_EXPORT_PLUGIN(NepomukAlignmentFactory("nepomukalignment"))

#include "nepomukalignment.moc"
