#include "kinotifydirwatch.h"
#include "kinotifywatch.h"
#include "kinotifywatchpool.h"

#include <QtCore/QSocketNotifier>
#include <QtCore/QHash>
#include <QtCore/QDir>

#include <kdebug.h>

#include <sys/inotify.h>
#include <unistd.h>
#include <fcntl.h>

static const int EVENT_STRUCT_SIZE = sizeof( struct inotify_event );

// we need one event to fit into the buffer, the problem is that the name
// is a variable length array
static const int EVENT_BUFFER_SIZE = EVENT_STRUCT_SIZE + 1024*16;

class KInotify::DirWatch::Private
{
public:
    Private( DirWatch* parent )
        : watchPool( 0 ),
          m_parent( parent ),
          m_inotifyFd( -1 ),
          m_notifier( 0 ) {
    }

    ~Private() {
        close();
    }

    WatchPool* watchPool;
    QHash<int, QString> cookies;

    unsigned char eventBuffer[EVENT_BUFFER_SIZE];

    int inotify() {
        if ( m_inotifyFd < 0 ) {
            open();
        }
        return m_inotifyFd;
    }

    void close() {
        kDebug();
        // We rely on the QObject hierarchy for cleanup
        delete watchPool;
        watchPool = 0;

        delete m_notifier;
        m_notifier = 0;

        ::close( m_inotifyFd );
        m_inotifyFd = -1;
    }

private:
    void open() {
        kDebug();
        m_inotifyFd = inotify_init();
        delete m_notifier;
        if ( m_inotifyFd > 0 ) {
            fcntl( m_inotifyFd, F_SETFD, FD_CLOEXEC );
            kDebug() << "Successfully opened connection to inotify:" << m_inotifyFd;
            m_notifier = new QSocketNotifier( m_inotifyFd, QSocketNotifier::Read );
            connect( m_notifier, SIGNAL( activated( int ) ), m_parent, SLOT( slotEvent( int ) ) );
            watchPool = new WatchPool( m_parent );
        }
    }

    DirWatch* m_parent;
    int m_inotifyFd;
    QSocketNotifier* m_notifier;
};


KInotify::DirWatch::DirWatch( QObject* parent )
    : QObject( parent ),
      d( new Private( this ) )
{
}


KInotify::DirWatch::~DirWatch()
{
    delete d;
}


bool KInotify::DirWatch::addWatch( const QString& path, WatchModes modes, WatchFlags flags, bool recursive )
{
    // FIXME: the path may not be unique.
    Watch* newWatch = new Watch( d->watchPool, d->watchPool, d->inotify(), path, modes, flags, recursive );
    if ( newWatch->isValid() ) {
        return true;
    }
    else {
        delete newWatch;
        return false;
    }
}


void KInotify::DirWatch::removeWatch( const QString& path )
{
    // FIXME: the path may not be unique.
    Watch* watch = d->watchPool->watchForPath( path );
    if ( watch ) {
        delete watch;
    }
}


bool KInotify::DirWatch::watchingPath( const QString& path )
{
    // FIXME: the path may not be unique.
    return d->watchPool->watchForPath( path );
}


void KInotify::DirWatch::slotEvent( int socket )
{
    // read at least one event
    int len = read( socket, d->eventBuffer, EVENT_BUFFER_SIZE );
    int i = 0;
    while ( i < len && len-i >= EVENT_STRUCT_SIZE  ) {
        struct inotify_event* event = ( struct inotify_event* )&d->eventBuffer[i];
        Watch* watch = d->watchPool->watchById( event->wd );

        // FIXME: do we need to prefix the dir?
        QString path = QFile::decodeName( QByteArray::fromRawData( event->name, event->len ) );
        if ( path[0] != '/' ) {
            // dir watches use relative paths
            QString p( watch->path() );
            if ( p[p.length()-1] != '/' ) {
                p += '/';
            }
            path.prepend( p );
        }

        // now signal the event
        if ( event->mask & EventAccess) {
            kDebug() << path << "EventAccess";
            emit accessed( path );
        }
        if ( event->mask & EventAttributeChange ) {
            kDebug() << path << "EventAttributeChange";
            emit attributeChanged( path );
        }
        if ( event->mask & EventCloseWrite ) {
            kDebug() << path << "EventCloseWrite";
            emit closedWrite( path );
        }
        if ( event->mask & EventCloseRead ) {
            kDebug() << path << "EventCloseRead";
            emit closedRead( path );
        }
        if ( event->mask & EventCreate ) {
            if ( watch->recursive() && ( event->mask & IN_ISDIR ) ) {
                watch->addSubWatch( path );
            }
            kDebug() << path << "EventCreate";
            emit created( path );
        }
        if ( event->mask & EventDelete ) {
            kDebug() << path << "EventDelete";
            emit deleted( path );
        }
        if ( event->mask & EventDeleteSelf ) {
            kDebug() << path << "EventDeleteSelf";
            emit deleted( watch->path() );
        }
        if ( event->mask & EventModify ) {
            kDebug() << path << "EventModify";
            emit modified( path );
        }
        if ( event->mask & EventMoveSelf ) {
            // FIXME: update the watch? Where do we get the new path? Watched folders are handled below to
            // a certain degree. At least when watching everything recursively we should not have a problem.
            kDebug() << watch->path() << "EventMoveSelf" << path;
//            emit moved( watch->path(), path );
        }
        if ( event->mask & EventMoveFrom ) {
            kDebug() << path << "EventMoveFrom";
            d->cookies[event->cookie] = path;
        }
        if ( event->mask & EventMoveTo ) {
            // check if ew have a cookie for this one
            if ( d->cookies.contains( event->cookie ) ) {
                QString oldPath = d->cookies[event->cookie];
                d->cookies.remove( event->cookie );
                kDebug() << oldPath << "EventMoveTo" << path;

                if ( event->mask & IN_ISDIR ) {
                    // check if we watch that dir and update the paths of the
                    // watch (excluding subwatches as they will have their own events) accordingly
                    if ( Watch* w = d->watchPool->watchForPath( oldPath ) ) {
                        w->updatePath( path );
                    }
                }

                emit moved( oldPath, path );
            }
            else {
                kDebug() << "No cookie for move information of" << path;
            }
        }
        if ( event->mask & EventOpen ) {
            kDebug() << path << "EventOpen";
            emit opened( path );
        }
        if ( event->mask & EventUnmount ) {
            kDebug() << path << "EventUnmount";
            emit unmounted( watch->path() );
        }
        if ( event->mask & EventQueueOverflow ) {
            kDebug() << path << "EventQueueOverflow";
            emit queueOverflow();
        }
        if ( event->mask & EventIgnored ) {
            // FIXME: should we remove the watch here?
            kDebug() << path << "EventIgnored";
        }

        // a one-shot watch will never fire again and, thus, is useless
        if ( watch->flags() & FlagOneShot ) {
            removeWatch( watch->path() );
        }

        i += EVENT_STRUCT_SIZE + event->len;
    }

    if ( len < 0 ) {
        kDebug() << "Failed to read event.";
    }
}

#include "kinotifydirwatch.moc"
