#include "kinotifywatch.h"
#include "kinotifywatchpool.h"

#include <QtCore/QFile>
#include <QtCore/QDir>

#include <kdebug.h>

#include <sys/inotify.h>
#include <errno.h>


class KInotify::Watch::Private
{
public:
    Private( Watch* parent )
        : wd( -1 ),
          recursive( false ),
          m_parent( parent ) {
    }
    ~Private() {
        if ( wd > 0 ) {
            inotify_rm_watch( inotifyId, wd );
        }
        // No need to delete sub-watches, QObject takes care of that.
    }

    QString path;
    WatchModes mode;
    WatchFlags flags;

    int inotifyId;
    int wd;

    bool recursive;

    QList<Watch*> subWatches;

    WatchPool* watchPool;

    void updateSubWatches();
    Watch* findSubWatch( const QString& path );
    Watch* createSubWatch( const QString& path );

private:
    Watch* m_parent;
};


void KInotify::Watch::Private::updateSubWatches()
{
    QDir dir( path );
    if ( dir.exists() ) {
        QStringList subDirs = dir.entryList( QDir::Dirs );
        subDirs.removeAll( "." );
        subDirs.removeAll( ".." );

        // add missing watches
        Q_FOREACH( QString s, subDirs ) {
            QString path = dir.absoluteFilePath( s );
            if ( !findSubWatch( path ) ) {
                createSubWatch( path );
            }
        }
    }
    else {
        kDebug() << "Cannot create sub watches for" << path;
    }
}


KInotify::Watch* KInotify::Watch::Private::createSubWatch( const QString& path )
{
    Watch* w = new Watch( m_parent, watchPool, inotifyId, path, mode, flags, true );
    subWatches.append( w );
    return w;
}


KInotify::Watch* KInotify::Watch::Private::findSubWatch( const QString& path )
{
    Q_FOREACH( Watch* watch, subWatches ) {
        if ( watch->path() == path ) {
            return watch;
        }
    }
    return 0;
}



KInotify::Watch::Watch( QObject* parent, WatchPool* pool,
                        int inotifyId, const QString& path,
                        WatchModes mode, WatchFlags flags, bool recursive )
    : QObject( parent ),
      d( new Private( this ) )
{
    int mask = mode|flags;
    d->wd = inotify_add_watch( inotifyId, QFile::encodeName( path ).data(), mask );
    if ( d->wd > 0 ) {
        kDebug() << "Successfully opened watch for" << path << ":" << d->wd;
        d->inotifyId = inotifyId;
        d->path = path;
        d->mode = mode;
        d->flags = flags;
        d->recursive = recursive;
        d->watchPool = pool;
        d->watchPool->registerWatch( this );
        if ( recursive ) {
            d->updateSubWatches();
        }
    }
    else {
        kDebug() << "failed to create watch for" << path;
        static bool userLimitReachedSignaled = false;
        if ( !userLimitReachedSignaled && errno == ENOSPC ) {
            userLimitReachedSignaled = true;
            emit watchUserLimitReached();
        }
    }
}


KInotify::Watch::~Watch()
{
    if ( Watch* watch = dynamic_cast<Watch*>( parent() ) ) {
        watch->subWatchRemoved( this );
    }
    delete d;
}


bool KInotify::Watch::isValid() const
{
    return d->wd > 0;
}


int KInotify::Watch::descriptor() const
{
    return d->wd;
}


QString KInotify::Watch::path() const
{
    return d->path;
}


KInotify::WatchModes KInotify::Watch::mode() const
{
    return d->mode;
}


KInotify::WatchFlags KInotify::Watch::flags() const
{
    return d->flags;
}


bool KInotify::Watch::recursive() const
{
    return d->recursive;
}


void KInotify::Watch::modify( WatchModes mode, WatchFlags flags )
{
    int mask = mode|flags;
    d->wd = inotify_add_watch( d->inotifyId, QFile::encodeName( d->path ), mask );
}


void KInotify::Watch::subWatchRemoved( Watch* subWatch )
{
    d->subWatches.removeAll( subWatch );
}


void KInotify::Watch::addSubWatch( const QString& path )
{
    d->createSubWatch( path );
}


void KInotify::Watch::updatePath( const QString& path )
{
    d->path = path;
}


// bool KInotify::Watch::watches( const QString& path )
// {

// }

#include "kinotifywatch.moc"
