/* This file is part of the KDE Project
   Copyright (c) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "nepomukfilewatch.h"
#include "kinotifydirwatch.h"

#include <QtCore/QTimer>
#include <QtCore/QDir>
#include <QtCore/QRegExp>
#include <QtCore/QFileInfo>

#include <kdebug.h>
#include <kurl.h>
#include <kmessagebox.h>
#include <klocale.h>

#include <Soprano/Model>
#include <Soprano/StatementIterator>
#include <Soprano/Statement>
#include <Soprano/Node>
#include <Soprano/NodeIterator>
#include <Soprano/QueryResultIterator>
#include <Soprano/Client/DBusClient>
#include <Soprano/Client/DBusModel>

using namespace Soprano;

namespace {
    Soprano::QueryResultIterator queryChildren( Model* model, const QString& path )
    {
        // escape special chars
        QString regexpPath( path );
        if ( regexpPath[regexpPath.length()-1] != '/' ) {
            regexpPath += '/';
        }
        regexpPath.replace( QRegExp( "([\\.\\?\\*\\\\+\\(\\)\\\\\\|\\[\\]{}])" ), "\\\\\\1" );

//        kDebug() << "query:" << QString( "select ?r ?p where { ?r <http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#fileUrl> ?p FILTER(REGEX(STR(?p), '^%1')) . }" ).arg( regexpPath );

        // query for all files that
        return model->executeQuery( QString( "prefix xesam: <http://freedesktop.org/standards/xesam/1.0/core#> "
                                             "select ?r ?p where { ?r xesam:url ?p FILTER(REGEX(STR(?p), '^%1')) . }" ).arg( regexpPath ),
                                    Soprano::Query::QueryLanguageSparql );
    }
}



Nepomuk::FileWatch::FileWatch( QObject* parent )
    : KDEDModule( parent ),
      m_locationUri("http://freedesktop.org/standards/xesam/1.0/core#url"),
      m_sopranoClient( 0 ),
      m_sopranoModel( 0 )
{
    m_dirWatch = new KInotify::DirWatch( this );
    connect( m_dirWatch, SIGNAL( moved( const QString&, const QString& ) ),
             this, SLOT( slotFileMoved( const QString&, const QString& ) ) );
    connect( m_dirWatch, SIGNAL( deleted( const QString& ) ),
             this, SLOT( slotFileDeleted( const QString& ) ) );
    connect( m_dirWatch, SIGNAL( watchUserLimitReached() ),
             this, SLOT( slotWatchUserLimitReached() ) );

//    QTimer::singleShot( 0, this, SLOT( updateWatches() ) );

    // recursively watch the whole home dir

    // FIXME: we cannot simply watch the folders that contain annotated files since moving
    // one of these files out of the watched "area" would mean we "loose" it, i.e. we have no
    // information about where it is moved.
    // On the other hand only using the homedir means a lot of restrictions.
    // One dummy solution would be a hybrid: watch the whole home dir plus all folders that
    // contain annotated files outside of the home dir and hope for the best

    m_dirWatch->addWatch( QDir::homePath(),
                          KInotify::WatchModes( KInotify::WatchMove|KInotify::WatchDelete|KInotify::WatchDeleteSelf ),
                          KInotify::WatchFlags(),
                          true );
}


Nepomuk::FileWatch::~FileWatch()
{
}


Soprano::Model* Nepomuk::FileWatch::sopranoModel()
{
    if ( !m_sopranoClient ) {
        m_sopranoClient = new Soprano::Client::DBusClient( "org.kde.NepomukServer", this );
    }
    if ( !m_sopranoModel ) {
        m_sopranoModel = m_sopranoClient->createModel( "main" );
    }
    return m_sopranoModel;
}


void Nepomuk::FileWatch::slotStatementsAdded()
{
    updateWatches();
}


void Nepomuk::FileWatch::updateWatches()
{
    // FIXME: handle files that are in an archive or virtual filesystem of some kind by for example
    // watching the archive file for change

    if ( sopranoModel() ) {
        Soprano::StatementIterator it = sopranoModel()->listStatements( Soprano::Statement( Soprano::Node(),
                                                                                            m_locationUri,
                                                                                            Soprano::Node() ) );

        while ( it.next() ) {
            Soprano::Node resource = it.current().subject();
            QString path = it.current().object().toString();

            // FIXME: would it be better to just watch the whole home dir?
//             if ( !m_dirWatch->watchingPath( path ) ) {
//                  m_dirWatch->addWatch( path );
//             }
        }
    }
    else {
        kDebug() << "Could not contact Nepomuk server.";
    }
}


void Nepomuk::FileWatch::slotFileMoved( const QString& from, const QString& to )
{
    if ( sopranoModel() ) {
        // We do NOT get deleted messages for overwritten files! Thus, we have to remove all metadata for overwritten files
        // first. We do that now.
        removeMetaData( to );

        // and finally update the old statements
        updateMetaData( from, to );

        // update children files (FIXME: just to be save let KInotify signal if it is a dir or not)
        foreach( Soprano::BindingSet bs, queryChildren( sopranoModel(), from ).allBindings() ) {
            QString path = to;
            if ( !to.endsWith( '/' ) )
                path += '/';
            path += bs[1].toString().mid( from.endsWith( '/' ) ? from.length() : from.length()+1 );
            updateMetaData( bs[1].toString(), path ); // FIXME: reuse the URI we already have
        }
    }
    else {
        kDebug() << "Could not contact Nepomuk server.";
    }
}


void Nepomuk::FileWatch::slotFileDeleted( const QString& path )
{
    if ( sopranoModel() ) {
        // just remove the metadata
        removeMetaData( path );

        // remove child annotations
        foreach( Soprano::Node node, queryChildren( sopranoModel(), path ).iterateBindings( 0 ).allNodes() ) {
            sopranoModel()->removeAllStatements( Statement( node, Node(), Node() ) );
        }
    }
    else {
        kDebug() << "Could not contact Nepomuk server.";
    }
}


void Nepomuk::FileWatch::removeMetaData( const QString& path )
{
    kDebug() << "remove" << path;

    // FIXME: folder paths may have an additional separator at the end. So maybe we need to test for both?
    QList<Statement> sl = sopranoModel()->listStatements( Statement( Node(),
                                                                     m_locationUri,
                                                                     LiteralValue( path ) ) ).allStatements();
    // the result should be zero or one statement... but who knows
    // TODO: this may be a little hardcore! Is there a milder way?
    Q_FOREACH ( Statement s, sl ) {
        kDebug() << "removing" << s.subject();
        sopranoModel()->removeAllStatements( Statement( s.subject(), Node(), Node() ) );
    }

    // FIXME: what about the triples that have our uri as object?
}


void Nepomuk::FileWatch::updateMetaData( const QString& from, const QString& to )
{
    kDebug() << from << "->" << to;

    // get statements to update
    QList<Soprano::Statement> sl = sopranoModel()->listStatements( Soprano::Statement( Soprano::Node(),
                                                                                       m_locationUri,
                                                                                       Soprano::LiteralValue( from ) ) ).allStatements();
    Soprano::Node resource;

    // update nfo:fileUri
    // -----------------------------------------------
    Q_FOREACH( Soprano::Statement s, sl ) {
        // update fileUrl
        kDebug() << "Updating" << s.subject();
        sopranoModel()->addStatement( Soprano::Statement( s.subject(),
                                                          s.predicate(),
                                                          Soprano::LiteralValue( to ),
                                                          s.context() ) );
        sopranoModel()->removeStatement( s );

        resource = s.subject();
    }


    // also change the resource URI to match the file URL
    // this is far from perfect but after all we use file URLs for resource URIs :(
    // -----------------------------------------------
    if ( resource.isResource() &&
         resource.uri() == KUrl::fromPathOrUrl( from ) ) {

        Soprano::Node newResource = KUrl::fromPathOrUrl( to );

        // update the resource itself
        // -----------------------------------------------
        sl = sopranoModel()->listStatements( Soprano::Statement( resource,
                                                                 Soprano::Node(),
                                                                 Soprano::Node() ) ).allStatements();
        Q_FOREACH( Soprano::Statement s, sl ) {
            sopranoModel()->addStatement( Soprano::Statement( newResource,
                                                              s.predicate(),
                                                              s.object(),
                                                              s.context() ) );

        }
        sopranoModel()->removeStatements( sl );
        // -----------------------------------------------


        // update resources relating to it
        // -----------------------------------------------
        sl = sopranoModel()->listStatements( Statement( Node(),
                                                        Node(),
                                                        resource ) ).allStatements();
        Q_FOREACH( Soprano::Statement s, sl ) {
            sopranoModel()->addStatement( Soprano::Statement( s.subject(),
                                                              s.predicate(),
                                                              newResource,
                                                              s.context() ) );

        }
        sopranoModel()->removeStatements( sl );
        // -----------------------------------------------
    }
}


void Nepomuk::FileWatch::slotWatchUserLimitReached()
{
    // inform the user
    KMessageBox::error( 0,
                        i18n( "<p>The Nepomuk file watch has reached the user limit of file watches. "
                              "That means new folders cannot be watched for changes anymore and "
                              "your metadata might be unaccessible after moving or renaming files. "
                              "You can try to increase the maximum allowed watches in:"
                              "<code>/proc/sys/fs/inotify/</code>" ),
                        i18n( "Nepomuk File Modification Watch Error" ) );
}


extern "C"
{
    KDE_EXPORT KDEDModule *create_nepomukfilewatch()
    {
        return new Nepomuk::FileWatch();
    }
}
