/*
   This file is part of the Nepomuk KDE project.
   Copyright (C) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
 */

#include "genericresourceguiitem.h"
#include "ratingpainter.h"

#include <QtGui/QPainter>
#include <QtGui/QFontMetrics>
#include <QtGui/QStyleOptionViewItem>
#include <QtGui/QMouseEvent>
#include <QtGui/QLinearGradient>

#include <KIcon>
#include <KColorScheme>
#include <KDebug>

#include <nepomuk/resource.h>
#include <nepomuk/tag.h>


namespace {
    QSize genericSize( Nepomuk::AbstractResourceGuiItem::DisplayMode mode )
    {
        switch( mode ) {
        case Nepomuk::AbstractResourceGuiItem::DisplaySimple:
            return QSize( 16, 16 );
        case Nepomuk::AbstractResourceGuiItem::DisplayOptimal:
            return QSize( 64, 64 );
        case Nepomuk::AbstractResourceGuiItem::DisplayFull:
            return QSize( 128, 128 );
        default:
            return QSize();
        }
    }

    QSize textSize( const QString& text, const QFontMetrics& metrics ) {
        return metrics.size( 0, text );
//         QSize size;
//         QStringList lines = text.split( '\n', QString::SkipEmptyParts );
//         for ( int i = 0; i < lines.count(); ++i ) {
//             int w = metrics.width( lines[i] );
//             size.setWidth( qMax( size.width(), w ) );
//         }
//         size.setHeight( lines.count() * metrics.height() + metrics.leading() * ( lines.count()-1 ) );
//         return size;
    }
}


class Nepomuk::GenericResourceGuiItem::Private
{
public:
    Private()
        : hoverRating( -1 ) {
    }

    // used for event caching
    QUrl currentlyChangedResource;
    int currentRating;
    int hoverRating;
};


Nepomuk::GenericResourceGuiItem::GenericResourceGuiItem( QObject* parent )
    : AbstractResourceGuiItem( parent ),
      d( new Private() )
{
}


Nepomuk::GenericResourceGuiItem::~GenericResourceGuiItem()
{
    delete d;
}


QVariant Nepomuk::GenericResourceGuiItem::data( const QUrl& resource, int role )
{
    Q_ASSERT( resource.isValid() );

    switch( role ) {
    case Qt::DecorationRole:
        return genericPixmap( resource, QSize( 16,16 ) );

    case Qt::DisplayRole:
        return genericLabel( resource );

    case Qt::ToolTipRole:
        return resource.toString();

    default:
        return QVariant();
    }
}


void Nepomuk::GenericResourceGuiItem::paint( QPainter* painter, DisplayMode mode, const QStyleOption& option, const QUrl& resource )
{
    Q_ASSERT( resource.isValid() );
    if ( !resource.isValid() ) {
        return;
    }

    Resource res( resource );

    QString label = genericLabel( res );
    QString description = genericDescription( res );
    int rating = res.rating();

    QList<Tag> tags = res.tags();

    QSize decorationSize = genericSize( mode );
    QStyleOptionViewItem::Position position = QStyleOptionViewItem::Left;
    Qt::Alignment align = Qt::AlignLeft|Qt::AlignVCenter;
    Qt::TextElideMode textElideMode = Qt::ElideRight; // FIXME: default to none and use a fading out effect
    QFontMetrics fontMetrics = option.fontMetrics;
    QFont descriptionFont = painter->font();
    descriptionFont.setPointSize( descriptionFont.pointSize()-2 );
    QFontMetrics descriptionFontMetrics( descriptionFont );
//    descriptionFont.setItalic( true );

    if ( option.type == QStyleOption::SO_ViewItem ) {
        const QStyleOptionViewItem& viewItemOption = ( const QStyleOptionViewItem& )option;
        decorationSize = viewItemOption.decorationSize;
        position = viewItemOption.decorationPosition;
//        align = viewItemOption.displayAlignment; // FIXME: So far we do not support another alignement than left since it looks ugly
        textElideMode = viewItemOption.textElideMode;
        fontMetrics = viewItemOption.font;
    }

    int spacing = 4;

    KColorScheme colorScheme( QPalette::Active, option.state & QStyle::State_Selected ? KColorScheme::Selection : KColorScheme::Window );
    QPixmap pix = genericPixmap( res, decorationSize );

    painter->save();

    if ( option.state & QStyle::State_Selected ) {
        painter->fillRect( option.rect, option.palette.highlight() );
    }

    switch( mode ) {
    case DisplaySimple: {
        QRect rect = option.rect;
        if ( 1 || position == QStyleOptionViewItem::Left ) {
            painter->drawPixmap( rect.topLeft(), pix );
            rect.setLeft( rect.left() + pix.width() + spacing );
        }
        else if ( position == QStyleOptionViewItem::Right ) {
            // FIXME: support other positions
        }

        painter->setPen( colorScheme.foreground( KColorScheme::NormalText ).color() );
        painter->drawText( rect, align, fontMetrics.elidedText( label, textElideMode, rect.width() ) );
        break;
    }

    case DisplayOptimal:
    case DisplayFull: {
        // FIXME: support other positions
        QRect rect = option.rect;
        painter->drawPixmap( rect.topLeft(), pix );
        rect.setLeft( rect.left() + pix.width() + spacing );

        painter->setPen( colorScheme.foreground( KColorScheme::NormalText ).color() );
        painter->drawText( rect, ( align^Qt::AlignVCenter )|Qt::AlignTop, fontMetrics.elidedText( label, textElideMode, rect.width() ) );
        rect.setTop( rect.top() + fontMetrics.height() + spacing );
        if ( rect.height() >= fontMetrics.height() ) {
            int flags = ( align&~( Qt::AlignVCenter|Qt::AlignBottom ) )|Qt::AlignTop|Qt::TextWordWrap;
            painter->save();
            // if the text does not fit entirely, fade it out at the bottom...
            if ( descriptionFontMetrics.boundingRect( QRect( 0, 0, rect.width(), 20000 ), flags, description ).height() > rect.height() ) {
                QLinearGradient gradient( QPoint( rect.left(), rect.bottom()-descriptionFontMetrics.height() ), rect.bottomLeft() );
                gradient.setColorAt( 0.0, colorScheme.foreground( KColorScheme::InactiveText ).color() );
                gradient.setColorAt( 1.0, colorScheme.background( KColorScheme::NormalBackground ).color() );
                QPen pen;
                pen.setBrush( gradient );
                painter->setPen( pen );
            }
            else {
                painter->setPen( colorScheme.foreground( KColorScheme::InactiveText ).color() );
            }
            painter->setFont( descriptionFont );
            painter->drawText( rect, flags, description );
            painter->restore();
        }

        QRect ratingRect = QRect( option.rect.x(),
                                  option.rect.y() + decorationSize.height() + spacing,
                                  decorationSize.width(),
                                  option.rect.height() - decorationSize.height() - spacing );

        RatingPainter::drawRating( painter,
                                   ratingRect,
                                   Qt::AlignTop,
                                   d->currentlyChangedResource == resource ? d->currentRating : rating,
                                   d->currentlyChangedResource == resource ? -1 : d->hoverRating );

        // tags in the lower right (max n chars, cut rest) (should be clickable) (make it a tag cloud in full mode!)
        break;
    }
    }

    painter->restore();
}


QSize Nepomuk::GenericResourceGuiItem::sizeHint( DisplayMode mode, const QStyleOption& option, const QUrl& resource )
{
    Q_ASSERT( resource.isValid() );

    Resource res( resource );

    QString label = genericLabel( res );
    QString description = genericDescription( res );

    QSize decorationSize = genericSize( mode );
    QStyleOptionViewItem::Position position = QStyleOptionViewItem::Left;
    QFontMetrics fontMetrics = option.fontMetrics;
    QFontMetrics descriptionFontMetrics  = fontMetrics;

    if ( option.type == QStyleOption::SO_ViewItem ) {
        const QStyleOptionViewItem& viewItemOption = ( const QStyleOptionViewItem& )option;
        decorationSize = viewItemOption.decorationSize;
        position = viewItemOption.decorationPosition;
        fontMetrics = viewItemOption.font;
        QFont descriptionFont = viewItemOption.font;
        descriptionFont.setPointSize( descriptionFont.pointSize()-2 );
        descriptionFontMetrics = descriptionFont;
    }

    int spacing = 4;

    switch( mode ) {
    case DisplaySimple:
        return QSize( decorationSize.width() + spacing + fontMetrics.width( label ),
                      qMax( decorationSize.height(), fontMetrics.height() ) );

    case DisplayOptimal:
//        if ( description.isEmpty() ) {
            return QSize( decorationSize.width() + spacing + fontMetrics.width( label ),
                          qMax( decorationSize.height(), fontMetrics.height() )
                          + decorationSize.width()/5 + spacing );
//         }
//         else {

//         }

    case DisplayFull:
        if ( description.isEmpty() ) {
            return QSize( decorationSize.width() + spacing + fontMetrics.width( label ),
                          qMax( decorationSize.height(), fontMetrics.height() )
                          + decorationSize.width()/5 + spacing );
        }
        else {
            QSize descriptionSize = textSize( description, descriptionFontMetrics );
            return QSize( decorationSize.width() + spacing + qMax( fontMetrics.width( label ), descriptionSize.width() ),
                          qMax( decorationSize.height(), fontMetrics.height() + spacing + descriptionSize.height() )
                          + decorationSize.width()/5 + spacing);
        }
    }

    return QSize();
}


QSize Nepomuk::GenericResourceGuiItem::minimumSizeHint( DisplayMode mode, const QStyleOption& option, const QUrl& resource )
{
    // FIXME: let the label be elided
    return sizeHint( mode, option, resource );
}


bool Nepomuk::GenericResourceGuiItem::handleEvent( QEvent* event, DisplayMode mode, const QStyleOption& option, const QUrl& resource )
{
    QSize decorationSize = genericSize( mode );
    QStyleOptionViewItem::Position position = QStyleOptionViewItem::Left;
    Qt::Alignment align = Qt::AlignLeft|Qt::AlignVCenter;
    Qt::TextElideMode textElideMode = Qt::ElideRight; // FIXME: default to none and use a fading out effect
    QFontMetrics fontMetrics = option.fontMetrics;

    if ( option.type == QStyleOption::SO_ViewItem ) {
        const QStyleOptionViewItem& viewItemOption = ( const QStyleOptionViewItem& )option;
        decorationSize = viewItemOption.decorationSize;
        position = viewItemOption.decorationPosition;
        align = viewItemOption.displayAlignment;
        textElideMode = viewItemOption.textElideMode;
        fontMetrics = viewItemOption.font;
    }

    int spacing = 4;

    if ( mode == DisplayFull || mode == DisplayOptimal ) {
        if ( event->type() == QEvent::MouseButtonPress ||
             event->type() == QEvent::MouseMove ||
             event->type() == QEvent::MouseButtonRelease ) {

            QRect ratingRect = QRect( option.rect.x(),
                                      option.rect.y() + decorationSize.height() + spacing,
                                      decorationSize.width(),
                                      option.rect.height() - decorationSize.height() - spacing );
            QMouseEvent* me = static_cast<QMouseEvent*>( event );

            switch( event->type() ) {
            case QEvent::MouseButtonPress:
                if ( me->button() == Qt::LeftButton ) {
                    d->currentRating = RatingPainter::getRatingFromPosition( ratingRect, Qt::AlignTop, option.direction, me->pos() );
                    if ( d->currentRating >= 0 ) {
                        d->currentlyChangedResource = resource;
                        return true;
                    }
                }
                break;

            case QEvent::MouseMove:
                d->hoverRating = RatingPainter::getRatingFromPosition( ratingRect, Qt::AlignTop, option.direction, me->pos() );
                if ( d->hoverRating >= 0 ) {
                    if ( d->currentlyChangedResource == resource ) {
                        d->currentRating = d->hoverRating;
                    }
                }
                return true;
                break;

            case QEvent::MouseButtonRelease:
                if ( d->currentlyChangedResource == resource ) {
                    d->currentlyChangedResource = QUrl(); // reset
                    int rating = RatingPainter::getRatingFromPosition( ratingRect, Qt::AlignTop, option.direction, me->pos() );
                    if ( rating >= 0 ) {
                        Resource res( resource );
                        res.setRating( rating );
                        return true;
                    }
                }
                break;

            default:
                // cannot happen but makes gcc happy
                return false;
            }
        }
    }

    return false;
}


QString Nepomuk::GenericResourceGuiItem::genericLabel( const Resource& resource )
{
    return resource.genericLabel();
}


QString Nepomuk::GenericResourceGuiItem::genericDescription( const Resource& resource )
{
    return resource.genericDescription();
}


QPixmap Nepomuk::GenericResourceGuiItem::genericPixmap( const Resource& resource, const QSize& size )
{
    QString icon = resource.genericIcon();
    if ( icon.isEmpty() ) {
        icon = "system-run"; // FIXME: use a better one!
    }
    return KIcon( icon ).pixmap( size );
}
