/*
   This file is part of the Nepomuk KDE project.
   Copyright (C) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
 */

#include "resourceguiitemfactory.h"
#include "genericresourceguiitem.h"

#include <KService>
#include <KServiceTypeTrader>
#include <KDebug>

#include <nepomuk/resourcemanager.h>
#include <nepomuk/class.h>

#include <Soprano/Model>
#include <Soprano/Statement>
#include <Soprano/StatementIterator>
#include <Soprano/Node>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/Xesam>



class Nepomuk::ResourceGuiItemFactory::Private
{
public:
    Private( ResourceGuiItemFactory* parent )
        : m_parent( parent ),
          m_genericItem( 0 ) {
    }

    AbstractResourceGuiItem* genericItem();
    AbstractResourceGuiItem* itemForType( const QUrl& type );
    void itemDestroyed( AbstractResourceGuiItem* item );

private:
    ResourceGuiItemFactory* m_parent;
    GenericResourceGuiItem* m_genericItem;

    QHash<QString, AbstractResourceGuiItem*> m_guiItemCache;
};


Nepomuk::AbstractResourceGuiItem* Nepomuk::ResourceGuiItemFactory::Private::genericItem()
{
    if ( !m_genericItem ) {
        m_genericItem = new GenericResourceGuiItem( m_parent );
        connect( m_genericItem, SIGNAL( destroyed( QObject* ) ),
                 m_parent, SLOT( slotGuiItemDestroyed( QObject* ) ) );
    }
    return m_genericItem;
}


Nepomuk::AbstractResourceGuiItem* Nepomuk::ResourceGuiItemFactory::Private::itemForType( const QUrl& type )
{
    if ( type.isEmpty() || type == Soprano::Vocabulary::RDFS::Resource() ) {
        return genericItem();
    }

    KService::List services = KServiceTypeTrader::self()->query( "Nepomuk/ResourceGuiItem",
                                                                 QString( "'%1' in [X-Nepomuk-ResourceTypes]" ).arg( type.toString() ) );
    if ( !services.isEmpty() ) {
        KSharedPtr<KService> service = services[0];
        if ( m_guiItemCache.contains( service->name() ) ) {
            return m_guiItemCache[service->name()];
        }
        else {
            if ( AbstractResourceGuiItem* item = service->createInstance<AbstractResourceGuiItem>( m_parent ) ) {
                m_guiItemCache.insert( service->name(), item );
                connect( item, SIGNAL( destroyed( QObject* ) ),
                         m_parent, SLOT( slotGuiItemDestroyed( QObject* ) ) );
                return item;
            }
        }
    }
    else {
        // search for superclasses!
        Nepomuk::Types::Class rdfClass( type );
        Q_FOREACH( Types::Class p, rdfClass.parentClasses() ) {
            if ( AbstractResourceGuiItem* item = itemForType( p.uri() ) ) {
                return item;
            }
        }
    }

    return 0;
}


void Nepomuk::ResourceGuiItemFactory::Private::itemDestroyed( AbstractResourceGuiItem* item )
{
    if ( item == m_genericItem ) {
        m_genericItem = 0;
    }
    else {
        m_guiItemCache.remove( m_guiItemCache.keys( item ).first() );
    }
}




Nepomuk::ResourceGuiItemFactory::ResourceGuiItemFactory( QObject* parent )
    : QObject( parent ),
      d( new Private( this ) )
{
}


Nepomuk::ResourceGuiItemFactory::~ResourceGuiItemFactory()
{
    delete d;
}


Nepomuk::AbstractResourceGuiItem* Nepomuk::ResourceGuiItemFactory::guiItemForType( const QUrl& type )
{
    if ( AbstractResourceGuiItem* item = d->itemForType( type ) ) {
        return item;
    }

    // fallback
    return d->genericItem();
}


Nepomuk::AbstractResourceGuiItem* Nepomuk::ResourceGuiItemFactory::guiItemForResource( const QUrl& resource )
{
    Soprano::StatementIterator it = ResourceManager::instance()->mainModel()->listStatements( resource,
                                                                                              Soprano::Vocabulary::RDF::type(),
                                                                                              Soprano::Node() );
    while ( it.next() ) {
        if ( AbstractResourceGuiItem* item = d->itemForType( it.current().object().uri() ) ) {
            return item;
        }
    }

    // fallback
    return d->genericItem();
}


// Nepomuk::AbstractResourceGuiItem* Nepomuk::ResourceGuiItemFactory::guiItemForMimeType( const QString& mimeType )
// {
//     // FIXME: how do we do this?

//     // fallback
//     return d->genericItem();
// }


void Nepomuk::ResourceGuiItemFactory::slotGuiItemDestroyed( QObject* item )
{
    d->itemDestroyed( static_cast<AbstractResourceGuiItem*>( item ) );
}

#include "resourceguiitemfactory.moc"
