/*
   This file is part of the Nepomuk KDE project.
   Copyright (C) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
 */

#include "queryparser.h"
#include "term.h"

#include <QtCore/QRegExp>

#include <KDebug>


namespace {
    QString s_plainTermPattern( "([^\\s\"':=\\<\\>]+|(?:([\"'])[^\"':=\\<\\>]+\\2))" );
    QString s_inExclusionPattern( "([\\+\\-]?)" );

    // match a simple search text
    // captures: 1 - The optional + or - sign (may be empty)
    //           2 - the search text (including optional paranthesis)
    QRegExp s_plainTermRx( "\\b" + s_inExclusionPattern + s_plainTermPattern + "\\b" );

    // match a field search term: fieldname + relation (:, =, etc) + search text with optional paranthesis
    // captures: 1 - The optional + or - sign (may be empty)
    //           2 - fieldname
    //           3 - relation
    //           4 - search text (including optional paranthesis)
    QRegExp s_fieldRx( "\\b" + s_inExclusionPattern + "(\\S+)([:=\\<\\>])" + s_plainTermPattern + "\\b" );

    Nepomuk::Search::Term::Type fieldTypeRelationFromString( const QString& s ) {
        if ( s == ":" ) {
            return Nepomuk::Search::Term::ContainsTerm;
        }
        else if ( s == "=" ) {
            return Nepomuk::Search::Term::EqualityTerm;
        }
        else {
            kDebug() << "FIXME: Unsupported relation:" << s;
            return Nepomuk::Search::Term::ContainsTerm;
        }
    }
}


Nepomuk::Search::Term Nepomuk::Search::QueryParser::parseQuery( const QString& query )
{
    // TODO: a "real" parser which can handle all of the Xesam user language
    //       This one for example does not handle nesting at all.

    QList<Term> terms;

    int pos = 0;
    while ( pos < query.length() ) {
        // skip whitespace
        while ( pos < query.length() && query[pos].isSpace() ) {
            kDebug() << "Skipping space at" << pos;
            ++pos;
        }

        if ( pos < query.length() ) {
            if ( s_fieldRx.indexIn( query, pos ) == pos ) {
                // FIXME: honour the +-
                kDebug() << "matched field term at" << pos << s_fieldRx.cap( 2 );
                Term term;
                term.setField( s_fieldRx.cap( 2 ) );
                term.setValue( s_fieldRx.cap( 4 ) );
                term.setType( fieldTypeRelationFromString( s_fieldRx.cap( 3 ) ) );
                terms.append( term );
                pos += s_fieldRx.matchedLength();
            }
            else if ( s_plainTermRx.indexIn( query, pos ) == pos ) {
                // FIXME: honour the +-
                kDebug() << "matched literal at" << pos << s_plainTermRx.cap( 2 );
                terms.append( Term( s_plainTermRx.cap( 2 ) ) );
                pos += s_plainTermRx.matchedLength();
            }
            else {
                kDebug() << "Invalid query at" << pos << query;
                return Term();
            }
        }
    }

    if ( terms.count() == 1 ) {
        return terms[0];
    }
    else if ( terms.count() > 0 ) {
        Term t;
        t.setType( Term::AndTerm );
        t.setSubTerms( terms );
        return t;
    }
    else {
        return Term();
    }
}


Nepomuk::Search::QueryParser::QueryParser()
{
}


Nepomuk::Search::QueryParser::~QueryParser()
{
}
