/*
   This file is part of the Nepomuk KDE project.
   Copyright (C) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
 */

#include "searchcore.h"
#include "searchthread.h"

#include <QtCore/QEventLoop>


class Nepomuk::Search::SearchCore::Private
{
public:
    Private()
        : maxResults( 20 ),
          cutOffScore( 0.25 ) {
    }

    int maxResults;
    double cutOffScore;
    QHash<QUrl, double> lastResults;
    SearchThread* searchThread;

    bool canceled;
};


Nepomuk::Search::SearchCore::SearchCore( QObject* parent )
    : QObject( parent ),
      d( new Private() )
{
    d->searchThread = new SearchThread( this );
    connect( d->searchThread, SIGNAL( newResult( const QUrl&, double ) ),
             this, SLOT( slotNewResult( const QUrl&, double ) ),
             Qt::QueuedConnection );
    connect( d->searchThread, SIGNAL( finished() ),
             this, SIGNAL( finished() ) );
}


Nepomuk::Search::SearchCore::~SearchCore()
{
    delete d;
}


void Nepomuk::Search::SearchCore::query( const Term& query )
{
    d->lastResults.clear();
    d->canceled = false;
    d->searchThread->query( query, cutOffScore(), maxResults() );
}


void Nepomuk::Search::SearchCore::cancel()
{
    d->canceled = true;
    d->searchThread->cancel();
}


void Nepomuk::Search::SearchCore::setMaxResults( int max )
{
    d->maxResults = max;
}


void Nepomuk::Search::SearchCore::setCutOffScore( double score )
{
    d->cutOffScore = qMax( 1.0, qMin( score, 0.0 ) );
}


int Nepomuk::Search::SearchCore::maxResults() const
{
    return d->maxResults;
}


double Nepomuk::Search::SearchCore::cutOffScore() const
{
    return d->cutOffScore;
}


QHash<QUrl, double> Nepomuk::Search::SearchCore::lastResults() const
{
    return d->lastResults;
}


void Nepomuk::Search::SearchCore::slotNewResult( const QUrl& resource, double score )
{
    if ( !d->canceled ) {
        QHash<QUrl, double>::iterator it = d->lastResults.find( resource );
        if ( it == d->lastResults.end() ) {
            d->lastResults.insert( resource, score );
            emit newResult( resource, score );
        }
        else {
            // FIXME: normalizing scores anyone??
            it.value() += score;
            emit scoreChanged( resource, it.value() );
        }
    }
}


QHash<QUrl, double> Nepomuk::Search::SearchCore::blockingQuery( const Term& query )
{
    SearchCore core;
    QEventLoop loop;
    connect( &core, SIGNAL( finished() ),
             &loop, SLOT( quit() ) );
    core.query( query );
    loop.exec();
    return core.lastResults();
}

#include "searchcore.moc"
