/*
   This file is part of the Nepomuk KDE project.
   Copyright (C) 2007 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
 */

#include "searchthread.h"
#include "term.h"

#include <nepomuk/resourcemanager.h>
#include <nepomuk/resource.h>

#include <Soprano/Client/DBusClient>
#include <Soprano/Client/DBusModel>
#include <Soprano/QueryResultIterator>
#include <Soprano/Node>
#include <Soprano/Statement>
#include <Soprano/LiteralValue>
#include <Soprano/StatementIterator>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/NRL>
#include <Soprano/Vocabulary/XMLSchema>

#include <KDebug>


// FIXME: With our cutoff score we might miss results that are hit multiple times and thus, would get their
//        score increased

uint qHash( const QUrl& url )
{
    return qHash( url.toString() );
}


using namespace Soprano;

namespace {
    void mergeInResult( QHash<QUrl, double>& results, const QUrl& resource, double score ) {
        QHash<QUrl, double>::iterator old = results.find( resource );
        if ( old == results.end() ) {
            results.insert( resource, score );
        }
        else {
            // FIXME: how do we join the scores properly? Is adding a good idea? It can certainly not be multiplication!
            *old += score;
        }
    }

    void mergeInResults( QHash<QUrl, double>& results, const QHash<QUrl, double>& otherResults ) {
        for ( QHash<QUrl, double>::const_iterator it = otherResults.constBegin();
              it != otherResults.constEnd(); ++it ) {
            mergeInResult( results, it.key(), it.value() );
        }
    }

    QString luceneQueryEscape( const QString& s ) {
        /* Chars to escape: + - && || ! ( ) { } [ ] ^ " ~ * ? : \ */

        static QRegExp rx( "([" + QRegExp::escape( "+-&|!(){}[]^\"~*?:\\" ) + "])" );
        QString es( s );
        es.replace( rx, "\\\\1" );
        return es;
    }

    QString createLiteralQuery( const QList<QUrl>& fields, const QString& value ) {
        QString newQuery;
        bool first = true;
        foreach( QUrl field, fields ) {
            if ( !first )
                newQuery += " OR ";
            newQuery += luceneQueryEscape( field.toString() ) + ":\"" + value + '\"';
            first = false;
        }
        return newQuery;
    }
}


Nepomuk::Search::SearchThread::SearchThread( QObject* parent )
    : QThread( parent )
{
    m_client = new Soprano::Client::DBusClient( "org.kde.NepomukServer", this );
    m_model = m_client->createModel( "main" );
}


Nepomuk::Search::SearchThread::~SearchThread()
{
    delete m_model;
}


void Nepomuk::Search::SearchThread::query( const Term& term, double cutOffScore, int maxResults )
{
    if( isRunning() ) {
        cancel();
    }

    m_canceled = false;
    m_searchTerm = term;
    m_cutOffScore = cutOffScore;
    m_maxResults = maxResults;
    m_numResults = 0;

    // FIXME: A bug in dbus prevents us from being multi-threaded.
    //        A patch exists but it has not been integrated into dbus yet
//    start();
    run();
}


void Nepomuk::Search::SearchThread::cancel()
{
    m_canceled = true;
    wait();
}


void Nepomuk::Search::SearchThread::run()
{
    search( m_searchTerm, 1.0, true );
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::search( const Term& term, double baseScore, bool reportResults )
{
    switch( term.type() ) {
    case Term::LiteralTerm:
        return literalSearch( term, baseScore, reportResults );

    case Term::AndTerm:
        return andSearch( term.subTerms(), baseScore, reportResults );

    case Term::OrTerm:
        return orSearch( term.subTerms(), baseScore, reportResults );

    case Term::ContainsTerm:
        return containsFieldSearch( term, baseScore, reportResults );

    case Term::EqualityTerm:
        return equalsFieldSearch( term, baseScore, reportResults );

    default:
        return QHash<QUrl, double>();
    }
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::andSearch( const QList<Term>& terms, double baseScore, bool reportResults )
{
    QHash<QUrl, double> results;
    bool first = true;
    foreach( Term term, terms ) {
        if ( m_canceled ) break;
        QHash<QUrl, double> termResults = search( term, baseScore, false );
        if ( first ) {
            results = termResults;
            first = false;
        }
        else {
            // intersect the results
            QHash<QUrl, double>::iterator it = results.begin();
            while ( it != results.end() ) {
                if ( m_canceled ) break;
                QHash<QUrl, double>::const_iterator termIt = termResults.find( it.key() );
                if ( termIt != termResults.constEnd() ) {
                    // update score
                    it.value() += termIt.value();
                    ++it;
                }
                else {
                    it = results.erase( it );
                }
            }
        }
    }

    if ( reportResults ) {
        for ( QHash<QUrl, double>::const_iterator it = results.constBegin();
              it != results.constEnd(); ++it ) {
            if ( m_canceled ) break;
            if ( m_numResults >= m_maxResults ) {
                return results;
            }
            else {
                ++m_numResults;
                emit newResult( it.key(), it.value() );
            }
        }
    }

    return results;
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::orSearch( const QList<Term>& terms, double baseScore, bool reportResults )
{
    QHash<QUrl, double> results;
    foreach( Term term, terms ) {
        if ( m_canceled ) break;
        mergeInResults( results, search( term, baseScore, reportResults ) );
    }
    if ( reportResults ) {
        for ( QHash<QUrl, double>::const_iterator it = results.constBegin();
              it != results.constEnd(); ++it ) {
            if ( m_canceled ) break;
            if ( m_numResults >= m_maxResults ) {
                return results;
            }
            else {
                ++m_numResults;
                emit newResult( it.key(), it.value() );
            }
        }
    }
    return results;
}


QList<QUrl> Nepomuk::Search::SearchThread::matchFieldName( const QString& field )
{
    kDebug() << field;

    QList<QUrl> results;

    // Step 1: see if we have a direct match to a predicate label
    QString query = QString( "select ?p where { "
                             "?p <%1> <%2> . "
                             "?p <%3> \"%4\"^^<%5> . }" )
                    .arg( Soprano::Vocabulary::RDF::type().toString() )
                    .arg( Soprano::Vocabulary::RDF::Property().toString() )
                    .arg( Soprano::Vocabulary::RDFS::label().toString() )
                    .arg( field )
                    .arg( Soprano::Vocabulary::XMLSchema::string().toString() );
    kDebug() << "Direct match query:" << query;

    Soprano::QueryResultIterator labelHits = m_model->executeQuery( query,
                                                                    Soprano::Query::QueryLanguageSparql );
    if ( !m_canceled ) {
        while ( labelHits.next() ) {
            results << labelHits.binding( "p" ).uri();
            kDebug() << "Found direct match" << labelHits.binding( "p" ).uri();
        }

        if ( results.isEmpty() ) {
            // FIXME: how about we have two repositories: one for the ontologies and one for the data.
            //        I don't think there will be relations between the RDF or Xesam ontology and some
            //        metadata....
            //        Because then queries like the one we are doing here will be more performant since
            //        we do not search the data itself and do not have to filter
            // BUT: What about inference?

            // FIXME: test which is better (performance-wise): Lucene query (here we have to do 2 steps) or SPARQL regexp
            query = QString( "select ?p where { "
                             "?p <%1> <%2> . "
                             "?p <%3> ?label . "
                             "FILTER(REGEX(STR(?label),'%4','i')) . }" )
                    .arg( Soprano::Vocabulary::RDF::type().toString() )
                    .arg( Soprano::Vocabulary::RDF::Property().toString() )
                    .arg( Soprano::Vocabulary::RDFS::label().toString() )
                    .arg( field );
            kDebug() << "Indirect hit query:" << query;
            labelHits = m_model->executeQuery( query,
                                               Soprano::Query::QueryLanguageSparql );
            QString newQuery;
            while ( labelHits.next() ) {
                results << labelHits.binding( "p" ).uri();
                kDebug() << "Found indirect match" << labelHits.binding( "p" ).uri();
            }
        }
    }

    return results;
}


/**
 * Searches resources that relate to \p resource.
 */
QHash<QUrl, double> Nepomuk::Search::SearchThread::relationSearch( const QUrl& resource, double baseScore, bool reportResults )
{
    kDebug() << resource << baseScore;

    QHash<QUrl, double> results;

    QueryResultIterator hits = m_model->executeQuery( QString( "select distinct ?r where { graph ?g { ?r ?p <%1> . } . ?g a <%2> . }" )
                                                      .arg( QString::fromAscii( resource.toEncoded() ) )
                                                      .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                                      Query::QueryLanguageSparql );

    while ( hits.next() ) {
        if ( m_canceled ) break;

        // We now equally distribute the score over all results.
        // TODO: this is where the relation weight graph from L3S would kick in
        results.insert( hits.binding( 0 ).uri(), baseScore );
        kDebug() << "relation result:" << hits.binding( 0 ).uri();
    }

    for ( QHash<QUrl, double>::iterator it = results.begin();
          it != results.end(); ++it ) {
        if ( m_canceled ) break;

        it.value() = baseScore / ( double )results.count();
        if ( reportResults ) {
            if ( m_numResults >= m_maxResults ) {
                break;
            }
            else {
                ++m_numResults;
                emit newResult( it.key(), it.value() );
            }
        }
    }

    return results;
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::equalsFieldSearch( const Term& query, double baseScore, bool reportResults )
{
    kDebug() << query.field() << query.value() << baseScore;

    Q_ASSERT( query.type() == Term::EqualityTerm );

    QHash<QUrl, double> results;

    // Step 1: get the proper properties from the field name
    // -----------------------------------------------------
    QList<QUrl> properties = matchFieldName( query.field() );

    foreach( QUrl property, properties ) {
        if ( m_canceled ) break;

        // Step 2: Look for direct hits
        // FIXME: handle more than just strings!
        // -----------------------------------------------------
        Soprano::QueryResultIterator hits = m_model->executeQuery( QString( "select distinct ?r where { "
                                                                            "graph ?g { ?r <%1> \"%2\"^^<%3> . } . "
                                                                            "?g a <%4> . }" )
                                                                   .arg( property.toString() )
                                                                   .arg( query.value() )
                                                                   .arg( Soprano::Vocabulary::XMLSchema::string().toString() )
                                                                   .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                                                   Soprano::Query::QueryLanguageSparql );
        while ( hits.next() ) {
            if ( m_canceled ) break;

            QUrl hitUri = hits.binding( 0 ).uri();

            // these are actual direct hits and we can report them right away
            if ( reportResults ) {
                if ( m_numResults >= m_maxResults ) {
                    return results;
                }
                else {
                    ++m_numResults;
                    emit newResult( hitUri, baseScore );
                }
            }

            results.insert( hitUri, baseScore );
        }

        if ( m_canceled ) return results;

        // Step 3: Look for resource -> resource relations
        // We do not restrict the graph of the related resource since it may very well be an ontology
        // resource (in case of type queries)
        // -----------------------------------------------------
        hits = m_model->executeQuery( QString( "select ?r where { "
                                               "graph ?g1 { ?r <%1> ?o . } . "
                                               "?o <%2> \"%3\"^^<%4> . "
                                               "?g1 a <%5> . }" )
                                      .arg( property.toString() )
                                      .arg( Soprano::Vocabulary::RDFS::label().toString() )
                                      .arg( query.value() )
                                      .arg( Soprano::Vocabulary::XMLSchema::string().toString() )
                                      .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                      Soprano::Query::QueryLanguageSparql );
        while ( hits.next() ) {
            if ( m_canceled ) break;

            QUrl hitUri = hits.binding( 0 ).uri();

            // FIXME: hitScore is a lucene score which is not normalized

            kDebug() << "Found result:" << hitUri;

            // We treat this as a direct hit
            if ( reportResults ) {
                if ( m_numResults >= m_maxResults ) {
                    return results;
                }
                else {
                    ++m_numResults;
                    emit newResult( hitUri, baseScore );
                }
            }

            results.insert( hitUri, baseScore );
        }
    }

    return results;
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::containsFieldSearch( const Term& query, double baseScore, bool reportResults )
{
    Q_ASSERT( query.type() == Term::ContainsTerm );

    QHash<QUrl, double> results;

    // Step 1: get the proper properties from the field name
    // -----------------------------------------------------
    QList<QUrl> properties = matchFieldName( query.field() );

    if ( !properties.isEmpty() && !m_canceled ) {
        // Step 2: try direct hits (the "normal" desktop search way: resource -> literal)
        // -----------------------------------------------------
        QString luceneQuery = createLiteralQuery( properties, query.value() );
        Soprano::QueryResultIterator hits = m_model->executeQuery( luceneQuery,
                                                                   Soprano::Query::QueryLanguageUser,
                                                                   "lucene" );
        while ( hits.next() ) {
            if ( m_canceled ) break;

            QUrl hitUri = hits.binding( 0 ).uri();
            double hitScore = hits.binding( 1 ).literal().toDouble() * baseScore;

            // we only want results from knowledgebases
            if ( !m_model->executeQuery( QString( "ask where { graph ?g { <%1> ?p ?o . } . { ?g <%2> <%3> . } . }" )
                                         .arg( QString::fromAscii( hitUri.toEncoded() ) )
                                         .arg( Soprano::Vocabulary::RDF::type().toString() )
                                         .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                         Soprano::Query::QueryLanguageSparql ).boolValue() ) {
                kDebug() << "Not an instancebase hit" << hitUri;
                continue;
            }

            if ( hitScore >= cutOffScore() ) {
                // these are actual direct hits and we can report them right away
                if ( reportResults ) {
                    if ( m_numResults >= m_maxResults ) {
                        return results;
                    }
                    else {
                        ++m_numResults;
                        emit newResult( hitUri, hitScore );
                    }
                }

                results.insert( hitUri, hitScore );
            }
        }

        if ( m_canceled ) return results;

        // Step 3: find resource -> resource relations by first searching
        //         the query value and then matching the resulting resources
        //         to our query field
        // -----------------------------------------------------
        hits = m_model->executeQuery( '\"' + query.value() + '\"',
                                      Soprano::Query::QueryLanguageUser,
                                      "lucene" );
        while ( hits.next() ) {
            if ( m_canceled ) break;

            QUrl hitUri = hits.binding( 0 ).uri();
            double hitScore = hits.binding( 1 ).literal().toDouble() * baseScore;

            // now for each of our matching predicates search the resources that
            // relate to the hit
            foreach( QUrl uri, properties ) {
                if ( m_canceled ) break;

                kDebug() << "Searching for hit" << hitUri << "as" << uri;
                QueryResultIterator it = m_model->executeQuery( QString( "select ?r where { graph ?g { ?r <%1> <%2> . } . ?g a <%3> . }" )
                                                                .arg( uri.toString() )
                                                                .arg( QString::fromAscii( hitUri.toEncoded() ) )
                                                                .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                                                Query::QueryLanguageSparql );

                while ( it.next() ) {
                    if ( m_canceled ) break;

                    // FIXME: hitScore is a lucene score which is not normalized

                    kDebug() << "Found result:" << it.binding( 0 ).uri();

                    if ( hitScore * baseScore >= cutOffScore() ) {
                        // We treat this as a direct hit
                        if ( reportResults ) {
                            if ( m_numResults >= m_maxResults ) {
                                return results;
                            }
                            else {
                                ++m_numResults;
                                emit newResult( it.binding( 0 ).uri(), hitScore * baseScore );
                            }
                        }

                        results.insert( it.binding( 0 ).uri(), hitScore * baseScore );
                    }
                }
            }
        }
    }

    return results;
}


/**
 * Performs a simple literal search as known from "normal" desktop search systems, i.e.
 * fields are matched to a literal values.
 */
QHash<QUrl, double> Nepomuk::Search::SearchThread::plainLiteralSearch( const QString& query, double baseScore, bool reportResults )
{
    QHash<QUrl, double> results;

    Soprano::QueryResultIterator hits = m_model->executeQuery( '\"' + query + '\"',
                                                               Soprano::Query::QueryLanguageUser,
                                                               "lucene" );
    while ( hits.next() ) {
        if ( m_canceled ) break;

        QUrl hitUri = hits.binding( 0 ).uri();
        double hitScore = hits.binding( 1 ).literal().toDouble() * baseScore;

        kDebug() << "Hit:" << hitUri << hitScore;

        // savety net
        if ( hitUri.isEmpty() )
            continue;

        // we only want results from knowledgebases
        if ( !m_model->executeQuery( QString( "ask where { graph ?g { <%1> ?p ?o . } . { ?g <%2> <%3> . } . }" )
                                     .arg( QString::fromAscii( hitUri.toEncoded() ) )
                                     .arg( Soprano::Vocabulary::RDF::type().toString() )
                                     .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() ),
                                     Soprano::Query::QueryLanguageSparql ).boolValue() ) {
            kDebug() << "Not an instancebase hit" << hitUri;
            continue;
        }

        if ( hitScore >= cutOffScore() ) {
            results.insert( hitUri, hitScore );
            kDebug() << "Found result:" << hitUri << hitScore;
            if ( reportResults ) {
                if ( m_numResults >= m_maxResults ) {
                    return results;
                }
                else {
                    ++m_numResults;
                    emit newResult( hitUri, hitScore );
                }
            }
        }
    }

    return results;
}


QHash<QUrl, double> Nepomuk::Search::SearchThread::literalSearch( const Term& query, double baseScore, bool reportResults )
{
    Q_ASSERT( query.type() == Term::LiteralTerm );

    kDebug() << query.value() << baseScore;

    QHash<QUrl, double> results = plainLiteralSearch( query.value(), baseScore, reportResults );

    QHash<QUrl, double> relatedResults;
    for ( QHash<QUrl, double>::const_iterator resultsIt = results.constBegin();
          resultsIt != results.constEnd(); ++resultsIt ) {

        if ( m_canceled ) break;

        QUrl hitUri = resultsIt.key();
        double hitScore = resultsIt.value();

        // search related resources (with a lower score)
        QHash<QUrl, double> relations = relationSearch( hitUri, hitScore*0.5, reportResults );
        mergeInResults( relatedResults, relations );

        // This is where we enter the recursion
        for ( QHash<QUrl, double>::const_iterator it = relations.constBegin();
              it != relations.constEnd(); ++it ) {

            if ( m_canceled ) break;

            QUrl relationUri = it.key();
            double relationScore = it.value();

            Nepomuk::Resource resource( relationUri );

            // We do not use Resource::genericLabel since we do not want to search for filenames
            QString label = resource.label();
            if ( relationScore >= cutOffScore() && !label.isEmpty() ) {
                mergeInResults( relatedResults, literalSearch( label, relationScore, reportResults ) );
            }
        }
    }
    mergeInResults( results, relatedResults );

    return results;
}


#include "searchthread.moc"
